stackcache

package
v1.3.1 Latest Latest
Warning

This package is not in the latest version of its module.

Go to latest
Published: Apr 5, 2026 License: MIT Imports: 4 Imported by: 0

Documentation

Index

Constants

This section is empty.

Variables

This section is empty.

Functions

func FuncName

func FuncName(path string) string

FuncName reduces a fully qualified function name to the function name

func PackageName

func PackageName(path string) string

PackageName reduces a fully qualified function name to the package name This function is from logrus internals.

Types

type StackCache

type StackCache interface {
	GetCaller() runtime.Frame
	GetStackFrames() []runtime.Frame
}

func New

func New(pcSearchOffset, pcSkip int, breakpointPackage string) StackCache

New creates a new stack cache for effectively traversing runtime frame stack. The traverser will start at pcOffset and move until not exited from runtime internal packages of the output library. pcSkip frames will be cut to avoid reporting the middleware layers.

Directories

Path Synopsis
internal

Jump to