Documentation
¶
Index ¶
Constants ¶
This section is empty.
Variables ¶
This section is empty.
Functions ¶
func PackageName ¶
PackageName reduces a fully qualified function name to the package name This function is from logrus internals.
Types ¶
type StackCache ¶
func New ¶
func New(pcSearchOffset, pcSkip int, breakpointPackage string) StackCache
New creates a new stack cache for effectively traversing runtime frame stack. The traverser will start at pcOffset and move until not exited from runtime internal packages of the output library. pcSkip frames will be cut to avoid reporting the middleware layers.
Click to show internal directories.
Click to hide internal directories.